@@ -20,10 +20,12 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
+from __future__ import print_function
 
 import pdb
 import bpython
 
+
 class BPdb(pdb.Pdb):
     """ PDB with BPython support. """
 
@@ -38,20 +40,17 @@ class BPdb(pdb.Pdb):
         self.intro = None
         pdb.Pdb.postloop(self)
 
-    ### cmd.Cmd commands
-
+    # cmd.Cmd commands
 
     def do_Bpython(self, arg):
         bpython.embed(self.curframe.f_locals, ['-i'])
 
-
     def help_Bpython(self):
-        print "B(python)"
-        print
-        print ("Invoke the bpython interpreter for this stack frame. To exit "
-               "bpython and return to a standard pdb press Ctrl-d")
+        print("B(python)")
+        print("")
+        print("Invoke the bpython interpreter for this stack frame. To exit "
+              "bpython and return to a standard pdb press Ctrl-d")
 
-
-    ### shortcuts
+    # shortcuts
     do_B = do_Bpython
     help_B = help_Bpython
