@@ -39,7 +39,7 @@
 # - Instead the suspend key exits the program
 # - View source doesn't work on windows unless you install the less program (From GnuUtils or Cygwin)
 
-from __future__ import division, with_statement
+from __future__ import division
 
 import platform
 import os
@@ -58,13 +58,13 @@ if platform.system() != 'Windows':
 import unicodedata
 import errno
 
-import locale
 from types import ModuleType
+from six.moves import range
 
 # These are used for syntax highlighting
 from pygments import format
 from pygments.formatters import TerminalFormatter
-from pygments.lexers import PythonLexer
+from bpython._py3compat import PythonLexer
 from pygments.token import Token
 from bpython.formatter import BPythonFormatter
 
@@ -72,7 +72,7 @@ from bpython.formatter import BPythonFormatter
 from bpython import importcompletion
 
 # This for config
-from bpython.config import Struct
+from bpython.config import Struct, getpreferredencoding
 
 # This for keys
 from bpython.keys import cli_key_dispatch as key_dispatch
@@ -84,7 +84,6 @@ from bpython.translations import _
 from bpython import repl
 from bpython._py3compat import py3
 from bpython.pager import page
-from bpython import autocomplete
 import bpython.args
 
 if not py3:
@@ -99,9 +98,6 @@ DO_RESIZE = False
 # ---
 
 
-def getpreferredencoding():
-    return locale.getpreferredencoding() or sys.getdefaultencoding()
-
 def calculate_screen_lines(tokens, width, cursor=0):
     """Given a stream of tokens and a screen width plus an optional
     initial cursor position, return the amount of needed lines on the
@@ -285,7 +281,7 @@ def make_colors(config):
     }
 
     if platform.system() == 'Windows':
-        c = dict(c.items() +
+        c = dict(c.items() |
             [
             ('K', 8),
             ('R', 9),
@@ -322,7 +318,7 @@ class CLIInteraction(repl.Interaction):
         return reply.lower() in (_('y'), _('yes'))
 
 
-    def notify(self, s, n=10):
+    def notify(self, s, n=10, wait_for_keypress=False):
         return self.statusbar.message(s, n)
 
     def file_prompt(self, s):
@@ -353,6 +349,13 @@ class CLIRepl(repl.Repl):
         if config.cli_suggestion_width <= 0 or config.cli_suggestion_width > 1:
             config.cli_suggestion_width = 0.8
 
+    def _get_cursor_offset(self):
+        return len(self.s) - self.cpos
+    def _set_cursor_offset(self, offset):
+        self.cpos = len(self.s) - offset
+    cursor_offset = property(_get_cursor_offset, _set_cursor_offset, None,
+                             "The cursor offset from the beginning of the line")
+
     def addstr(self, s):
         """Add a string to the current input line and figure out
         where it should go, depending on the cursor position."""
@@ -430,7 +433,7 @@ class CLIRepl(repl.Repl):
             self.print_line(self.s)
 
     def clear_current_line(self):
-        """Called when a SyntaxError occured in the interpreter. It is
+        """Called when a SyntaxError occurred in the interpreter. It is
         used to prevent autoindentation from occuring after a
         traceback."""
         repl.Repl.clear_current_line(self)
@@ -441,39 +444,36 @@ class CLIRepl(repl.Repl):
         # curses does not handle this on its own. Sad.
         height, width = self.scr.getmaxyx()
         max_y = min(self.iy + (self.ix + len(self.s)) // width + 1, height)
-        for y in xrange(self.iy + 1, max_y):
+        for y in range(self.iy + 1, max_y):
             self.scr.move(y, 0)
             self.scr.clrtoeol()
 
     def complete(self, tab=False):
-        """Get Autcomplete list and window."""
-        if self.paste_mode and self.list_win_visible:
-            self.scr.touchwin()
+        """Get Autcomplete list and window.
 
+        Called whenever these should be updated, and called
+        with tab
+        """
         if self.paste_mode:
+            self.scr.touchwin() #TODO necessary?
             return
 
-        if self.list_win_visible and not self.config.auto_display_list:
-            self.scr.touchwin()
-            self.list_win_visible = False
-            self.matches_iter.update()
-            return
-
-        if self.config.auto_display_list or tab:
-            self.list_win_visible = repl.Repl.complete(self, tab)
-            if self.list_win_visible:
-                try:
-                    self.show_list(self.matches, self.argspec)
-                except curses.error:
-                    # XXX: This is a massive hack, it will go away when I get
-                    # cusswords into a good enough state that we can start
-                    # using it.
-                    self.list_win.border()
-                    self.list_win.refresh()
-                    self.list_win_visible = False
-            if not self.list_win_visible:
-                self.scr.redrawwin()
-                self.scr.refresh()
+        list_win_visible = repl.Repl.complete(self, tab)
+        if list_win_visible:
+            try:
+                self.show_list(self.matches_iter.matches, self.arg_pos,
+                               topline=self.funcprops,
+                               formatter=self.matches_iter.completer.format)
+            except curses.error:
+                # XXX: This is a massive hack, it will go away when I get
+                # cusswords into a good enough state that we can start
+                # using it.
+                self.list_win.border()
+                self.list_win.refresh()
+                list_win_visible = False
+        if not list_win_visible:
+            self.scr.redrawwin()
+            self.scr.refresh()
 
     def clrtobol(self):
         """Clear from cursor to beginning of line; usual C-u behaviour"""
@@ -488,9 +488,12 @@ class CLIRepl(repl.Repl):
         self.scr.redrawwin()
         self.scr.refresh()
 
-    def current_line(self):
-        """Return the current line."""
+    def _get_current_line(self):
         return self.s
+    def _set_current_line(self, line):
+        self.s = line
+    current_line = property(_get_current_line, _set_current_line, None,
+                            "The characters of the current line")
 
     def cut_to_buffer(self):
         """Clear from cursor to end of line, placing into cut buffer"""
@@ -501,31 +504,6 @@ class CLIRepl(repl.Repl):
         self.scr.redrawwin()
         self.scr.refresh()
 
-    def cw(self):
-        """Return the current word, i.e. the (incomplete) word directly to the
-        left of the cursor"""
-
-        # I don't know if autocomplete should be disabled if the cursor
-        # isn't at the end of the line, but that's what this does for now.
-        if self.cpos: return
-
-        # look from right to left for a bad method or dictionary character
-        l = len(self.s)
-        is_method_char = lambda c: c.isalnum() or c in ('.', '_')
-        dict_chars = ['[']
-
-        if not self.s or not (is_method_char(self.s[-1])
-                                or self.s[-1] in dict_chars):
-            return
-
-        for i in range(1, l+1):
-            c = self.s[-i]
-            if not (is_method_char(c) or c in dict_chars):
-                i -= 1
-                break
-
-        return self.s[-i:]
-
     def delete(self):
         """Process a del"""
         if not self.s:
@@ -684,7 +662,7 @@ class CLIRepl(repl.Repl):
         self.iy, self.ix = self.scr.getyx()
 
         if not self.paste_mode:
-            for _ in xrange(self.next_indentation()):
+            for _ in range(self.next_indentation()):
                 self.p_key('\t')
 
         self.cpos = 0
@@ -715,7 +693,7 @@ class CLIRepl(repl.Repl):
         self.print_line(self.s, newline=True)
         self.echo("\n")
 
-    def mkargspec(self, topline, down):
+    def mkargspec(self, topline, in_arg, down):
         """This figures out what to do with the argspec and puts it nicely into
         the list window. It returns the number of lines used to display the
         argspec.  It's also kind of messy due to it having to call so many
@@ -723,16 +701,15 @@ class CLIRepl(repl.Repl):
         sturdy."""
 
         r = 3
-        fn = topline[0]
-        args = topline[1][0]
-        kwargs = topline[1][3]
-        _args = topline[1][1]
-        _kwargs = topline[1][2]
-        is_bound_method = topline[2]
-        in_arg = topline[3]
+        fn = topline.func
+        args = topline.argspec.args
+        kwargs = topline.argspec.defaults
+        _args = topline.argspec.varargs
+        _kwargs = topline.argspec.varkwargs
+        is_bound_method = topline.is_bound_method
         if py3:
-            kwonly = topline[1][4]
-            kwonly_defaults = topline[1][5] or dict()
+            kwonly = topline.argspec.kwonly
+            kwonly_defaults = topline.argspec.kwonly_defaults or dict()
         max_w = int(self.scr.getmaxyx()[1] * 0.6)
         self.list_win.erase()
         self.list_win.resize(3, max_w)
@@ -899,7 +876,9 @@ class CLIRepl(repl.Repl):
             return ''
 
         elif key in key_dispatch[config.undo_key]:  # C-r
-            self.undo()
+            n = self.prompt_undo()
+            if n > 0:
+                self.undo(n=n)
             return ''
 
         elif key in key_dispatch[config.search_key]:
@@ -983,19 +962,24 @@ class CLIRepl(repl.Repl):
             self.pastebin()
             return ''
 
+        elif key in key_dispatch[config.copy_clipboard_key]:
+            self.copy2clipboard()
+            return ''
+
         elif key in key_dispatch[config.last_output_key]:
             page(self.stdout_hist[self.prev_block_finished:-4])
             return ''
 
         elif key in key_dispatch[config.show_source_key]:
-            source = self.get_source_of_current_name()
-            if source is not None:
+            try:
+                source = self.get_source_of_current_name()
+            except repl.SourceNotFound as e:
+                self.statusbar.message(_(e))
+            else:
                 if config.highlight_show_source:
                     source = format(PythonLexer().get_tokens(source),
                                     TerminalFormatter())
                 page(source)
-            else:
-                self.statusbar.message(_('Cannot show source.'))
             return ''
 
         elif key in ('\n', '\r', 'PADENTER'):
@@ -1019,6 +1003,9 @@ class CLIRepl(repl.Repl):
         elif key == '\x18':
             return self.send_current_line_to_editor()
 
+        elif key == '\x03':
+            raise KeyboardInterrupt()
+
         elif key[0:3] == 'PAD' and not key in ('PAD0', 'PADSTOP'):
             pad_keys = {
                 'PADMINUS': '-',
@@ -1094,7 +1081,7 @@ class CLIRepl(repl.Repl):
         curses.raw(False)
         try:
             return repl.Repl.push(self, s, insert_into_history)
-        except SystemExit, e:
+        except SystemExit as e:
             # Avoid a traceback on e.g. quit()
             self.do_exit = True
             self.exit_value = e.args
@@ -1170,7 +1157,7 @@ class CLIRepl(repl.Repl):
 
         real_lineno = self.iy
         height, width = self.scr.getmaxyx()
-        for i in xrange(lineno, len(self.buffer)):
+        for i in range(lineno, len(self.buffer)):
             string = self.buffer[i]
             # 4 = length of prompt
             length = len(string.encode(getpreferredencoding())) + 4
@@ -1237,7 +1224,7 @@ class CLIRepl(repl.Repl):
         self.scr.refresh()
 
         if self.buffer:
-            for _ in xrange(indent):
+            for _ in range(indent):
                 self.tab()
 
         self.evaluating = False
@@ -1267,7 +1254,8 @@ class CLIRepl(repl.Repl):
         self.s_hist.append(s.rstrip())
 
 
-    def show_list(self, items, topline=None, current_item=None):
+    def show_list(self, items, arg_pos, topline=None, formatter=None, current_item=None):
+
         shared = Struct()
         shared.cols = 0
         shared.rows = 0
@@ -1283,23 +1271,12 @@ class CLIRepl(repl.Repl):
         self.list_win.erase()
 
         if items:
-            sep = '.'
-            separators = ['.', os.path.sep, '[']
-            lastindex = max([items[0].rfind(c) for c in separators])
-            if lastindex > -1:
-                sep = items[0][lastindex]
-            items = [x.rstrip(sep).rsplit(sep)[-1] for x in items]
+            items = [formatter(x) for x in items]
             if current_item:
-                current_item = current_item.rstrip(sep).rsplit(sep)[-1]
-
-            if items[0].endswith(']'):
-                # dictionary key suggestions
-                items = [x.rstrip(']') for x in items]
-                if current_item:
-                    current_item = current_item.rstrip(']')
+                current_item = formatter(current_item)
 
         if topline:
-            height_offset = self.mkargspec(topline, down) + 1
+            height_offset = self.mkargspec(topline, arg_pos, down) + 1
         else:
             height_offset = 0
 
@@ -1445,8 +1422,6 @@ class CLIRepl(repl.Repl):
         and don't indent if there are only whitespace in the line.
         """
 
-        mode = self.config.autocomplete_mode
-
         # 1. check if we should add a tab character
         if self.atbol() and not back:
             x_pos = len(self.s) - self.cpos
@@ -1458,66 +1433,40 @@ class CLIRepl(repl.Repl):
             self.print_line(self.s)
             return True
 
-        # 2. get the current word
+        # 2. run complete() if we aren't already iterating through matches
         if not self.matches_iter:
             self.complete(tab=True)
-            if not self.config.auto_display_list and not self.list_win_visible:
-                return True
-
-            cw = self.current_string() or self.cw()
-            if not cw:
-                return True
-        else:
-            cw = self.matches_iter.current_word
+            self.print_line(self.s)
 
         # 3. check to see if we can expand the current word
-        cseq = None
-        if mode == autocomplete.SUBSTRING:
-            if all([len(match.split(cw)) == 2 for match in self.matches]):
-                seq = [cw + match.split(cw)[1] for match in self.matches]
-                cseq = os.path.commonprefix(seq)
-        else:
-            seq = self.matches
-            cseq = os.path.commonprefix(seq)
-
-        if cseq and mode != autocomplete.FUZZY:
-            expanded_string = cseq[len(cw):]
-            self.s += expanded_string
-            expanded = bool(expanded_string)
+        if self.matches_iter.is_cseq():
+            #TODO resolve this error-prone situation:
+            # can't assign at same time to self.s and self.cursor_offset
+            # because for cursor_offset
+            # property to work correctly, self.s must already be set
+            temp_cursor_offset, self.s = self.matches_iter.substitute_cseq()
+            self.cursor_offset = temp_cursor_offset
             self.print_line(self.s)
-            if len(self.matches) == 1 and self.config.auto_display_list:
-                self.scr.touchwin()
-            if expanded:
-                self.matches_iter.update(cseq, self.matches)
-        else:
-            expanded = False
+            if not self.matches_iter:
+                self.complete()
 
         # 4. swap current word for a match list item
-        if not expanded and self.matches:
-            # reset s if this is the nth result
-            if self.matches_iter:
-                self.s = self.s[:-len(self.matches_iter.current())] + cw
-
+        elif self.matches_iter.matches:
             current_match = back and self.matches_iter.previous() \
-                                  or self.matches_iter.next()
-
-            # update s with the new match
-            if current_match:
-                try:
-                    self.show_list(self.matches, self.argspec, current_match)
-                except curses.error:
-                    # XXX: This is a massive hack, it will go away when I get
-                    # cusswords into a good enough state that we can start
-                    # using it.
-                    self.list_win.border()
-                    self.list_win.refresh()
-
-                if self.config.autocomplete_mode == autocomplete.SIMPLE:
-                    self.s += current_match[len(cw):]
-                else:
-                    self.s = self.s[:-len(cw)] + current_match
-
-                self.print_line(self.s, True)
+                                  or next(self.matches_iter)
+            try:
+                self.show_list(self.matches_iter.matches, self.arg_pos,
+                               topline=self.funcprops,
+                               formatter=self.matches_iter.completer.format,
+                               current_item=current_match)
+            except curses.error:
+                # XXX: This is a massive hack, it will go away when I get
+                # cusswords into a good enough state that we can start
+                # using it.
+                self.list_win.border()
+                self.list_win.refresh()
+            _, self.s = self.matches_iter.cur_line()
+            self.print_line(self.s, True)
         return True
 
     def undo(self, n=1):
@@ -1570,7 +1519,7 @@ class CLIRepl(repl.Repl):
         self.scr.refresh()
 
         if self.buffer:
-            for _ in xrange(indent):
+            for _ in range(indent):
                 self.tab()
 
         self.print_line(self.s)
@@ -1744,12 +1693,19 @@ def init_wins(scr, config):
     # Thanks to Angus Gibson for pointing out this missing line which was causing
     # problems that needed dirty hackery to fix. :)
 
-    statusbar = Statusbar(scr, main_win, background, config,
-        _(" <%s> Rewind  <%s> Save  <%s> Pastebin "
-          " <%s> Pager  <%s> Show Source ") %
-          (config.undo_key, config.save_key, config.pastebin_key,
-           config.last_output_key, config.show_source_key),
-            get_colpair(config, 'main'))
+    commands = (
+        (_('Rewind'), config.undo_key),
+        (_('Save'), config.save_key),
+        (_('Pastebin'), config.pastebin_key),
+        (_('Pager'), config.last_output_key),
+        (_('Show Source'), config.show_source_key)
+    )
+
+    message = '  '.join('<%s> %s' % (key, command) for command, key in commands
+                        if key)
+
+    statusbar = Statusbar(scr, main_win, background, config, message,
+                          get_colpair(config, 'main'))
 
     return main_win, statusbar
 
@@ -1951,7 +1907,7 @@ def main_curses(scr, args, config, interactive=True, locals_=None,
         exit_value = ()
         try:
             bpython.args.exec_code(interpreter, args)
-        except SystemExit, e:
+        except SystemExit as e:
             # The documentation of code.InteractiveInterpreter.runcode claims
             # that it reraises SystemExit. However, I can't manage to trigger
             # that. To be one the safe side let's catch SystemExit here anyway.
@@ -1961,7 +1917,12 @@ def main_curses(scr, args, config, interactive=True, locals_=None,
             return (exit_value, clirepl.getstdout())
     else:
         sys.path.insert(0, '')
-        clirepl.startup()
+        try:
+            clirepl.startup()
+        except OSError as e:
+            # Handle this with a proper error message.
+            if e.errno != errno.ENOENT:
+                raise
 
     if banner is not None:
         clirepl.write(banner)
@@ -2013,7 +1974,6 @@ def main(args=None, locals_=None, banner=None):
     return repl.extract_exit_value(exit_value)
 
 if __name__ == '__main__':
-    from bpython.cli import main
     sys.exit(main())
 
 # vim: sw=4 ts=4 sts=4 ai et
