@@ -5,7 +5,8 @@ import struct
 import sys
 import termios
 import textwrap
-import unittest
+
+from bpython.test import unittest
 
 try:
     from twisted.internet import reactor
@@ -13,14 +14,32 @@ try:
     from twisted.internet.protocol import ProcessProtocol
     from twisted.trial.unittest import TestCase as TrialTestCase
 except ImportError:
+    class TrialTestCase(object):
+        pass
     reactor = None
 
+try:
+    import urwid
+    have_urwid = True
+except ImportError:
+    have_urwid = False
+
+try:
+    from nose.plugins.attrib import attr
+except ImportError:
+    def attr(*args, **kwargs):
+        def identity(func):
+            return func
+        return identity
+
 TEST_CONFIG = os.path.join(os.path.dirname(__file__), "test.config")
 
+
 def set_win_size(fd, rows, columns):
     s = struct.pack('HHHH', rows, columns, 0, 0)
     fcntl.ioctl(fd, termios.TIOCSWINSZ, s)
 
+
 class CrashersTest(object):
     backend = "cli"
 
@@ -35,13 +54,13 @@ class CrashersTest(object):
         result = Deferred()
 
         class Protocol(ProcessProtocol):
-            STATES = (SEND_INPUT, COLLECT) = xrange(2)
+            STATES = (SEND_INPUT, COLLECT) = range(2)
 
             def __init__(self):
                 self.data = ""
                 self.delayed_call = None
                 self.states = iter(self.STATES)
-                self.state = self.states.next()
+                self.state = next(self.states)
 
             def outReceived(self, data):
                 self.data += data
@@ -56,7 +75,7 @@ class CrashersTest(object):
                     if index >= 0:
                         self.data = self.data[index + 4:]
                         self.transport.write(input)
-                        self.state = self.states.next()
+                        self.state = next(self.states)
                 else:
                     self.transport.closeStdin()
                     if self.transport.pid is not None:
@@ -70,13 +89,15 @@ class CrashersTest(object):
 
         (master, slave) = pty.openpty()
         set_win_size(slave, 25, 80)
-        reactor.spawnProcess(Protocol(), sys.executable,
-            (sys.executable, "-m", "bpython." + self.backend,
-             "--config",  TEST_CONFIG),
+        reactor.spawnProcess(
+            Protocol(), sys.executable,
+            (sys.executable, "-m", "bpython." + self.backend, "--config",
+             TEST_CONFIG),
             env=dict(TERM="vt100", LANG=os.environ.get("LANG", "")),
             usePTY=(master, slave, os.ttyname(slave)))
         return result
 
+    @attr(speed='slow')
     def test_issue108(self):
         input = textwrap.dedent(
             """\
@@ -87,6 +108,7 @@ class CrashersTest(object):
         deferred = self.run_bpython(input)
         return deferred.addCallback(self.check_no_traceback)
 
+    @attr(speed='slow')
     def test_issue133(self):
         input = textwrap.dedent(
             """\
@@ -97,15 +119,19 @@ class CrashersTest(object):
         return self.run_bpython(input).addCallback(self.check_no_traceback)
 
     def check_no_traceback(self, data):
-        tb = data[data.find("Traceback"):]
-        self.assertTrue("Traceback" not in data, tb)
+        self.assertNotIn("Traceback", data)
 
-if reactor is not None:
-    class CursesCrashersTest(TrialTestCase, CrashersTest):
-        backend = "cli"
 
-    class UrwidCrashersTest(TrialTestCase, CrashersTest):
-        backend = "urwid"
+@unittest.skipIf(reactor is None, "twisted is not available")
+class CursesCrashersTest(TrialTestCase, CrashersTest):
+    backend = "cli"
+
+
+@unittest.skipUnless(have_urwid, "urwid is required")
+@unittest.skipIf(reactor is None, "twisted is not available")
+class UrwidCrashersTest(TrialTestCase, CrashersTest):
+    backend = "urwid"
+
 
 if __name__ == "__main__":
     unittest.main()
